#define _BSD_SOURCE
#define _DEFAULT_SOURCE
#define _POSIX_C_SOURCE 199309L
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <time.h>
#include <sched.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <pcilib.h>
#include <pcilib/bar.h>
#include <pcilib/kmem.h>

#define DEVICE "/dev/fpga0"
#define BAR PCILIB_BAR0


#define WR(addr, value) { *(uint32_t*)(bar + addr) = value; }
#define RD(addr, value) { value = *(uint32_t*)(bar + addr); }

void request(void* volatile ua) {
    int fd = open("/sys/class/test/test0/test_request",  O_RDWR);
    if (fd >= 0) {
	size_t bytes;
	char res[64];

	sprintf(res, "%p", ua);
	write(fd, res, strlen(res));
	lseek(fd, SEEK_SET, 0);
	bytes = read(fd, res, sizeof(res) - 1); 

	printf("Res : ");
	if (bytes < 0) {
	    puts("Error");
	} else {
	    res[bytes] = 0;
	    puts(res);
	}

	close(fd);
    } else {
        printf("Error openning /sys/class/test/test0/test_request, errno %i\n", errno);
    }
}

int main(int argc, char *argv[]) {
    uint32_t i, j;
    pcilib_t *pci;
    pcilib_kmem_handle_t *kbuf;

    void* volatile bar;
    const pcilib_bar_info_t *bar_info;

    pci = pcilib_open(DEVICE, PCILIB_MODEL_DETECT);
    if (!pci) {
	printf("pcilib_open\n");
	exit(1);
    }

    kbuf = pcilib_alloc_kernel_memory(pci, PCILIB_KMEM_TYPE_PAGE, 1, 4096, 4096, 0, 0);
    void* volatile ua = pcilib_kmem_get_ua(pci, kbuf);
    uintptr_t pa = pcilib_kmem_get_pa(pci, kbuf);

    printf("Kmem\n-----\n");
    printf("User: %p\n", ua);
    printf("HW  : 0x%lx\n", pa);
    request(ua);
    printf("\n");

    bar = pcilib_resolve_bar_address(pci, BAR, 0);
    if (!bar) {
        pcilib_close(pci);
        printf("map bar\n");
        exit(1);
    }
    bar_info = pcilib_get_bar_info(pci, BAR);

    printf("Bar\n-----\n");
    printf("User: %p\n", bar_info[BAR].virt_addr);
    printf("HW  : %p\n", bar_info[BAR].phys_addr);
    request(bar_info[BAR].virt_addr);
    printf("\n");

    pcilib_free_kernel_memory(pci, kbuf, 0);
    pcilib_close(pci);
}
